from numpy import array
from thunder.images import fromtif

def load(path, nplanes, engine=None):
    data = fromtif(path, nplanes=nplanes, engine=engine)
    return data

def split(oim, nrois, nlines):
    """
    Splits a single two dimensional plane.
    """
    if not oim.ndim == 2:
      raise Exception('original image must be two dimensional')
    
    if nrois > 1:
        ntotal = oim.shape[0]
        ngap = (ntotal - nlines*nrois)/(nrois-1)
        return array([oim[int(nlines*i + ngap*i): int(nlines*(i + 1) + ngap*i), :] for i in range(nrois)])
    else:
        return oim

def merge(oim, order):
    """
    Merges a three dimensional volume into a single plane.
    """
    if not type(order) == bool:
        if not oim.ndim == 3:
          raise Exception('original image must be three dimensional')
        if not len(order) == oim.shape[0]:
          raise Exception('length of order must match first image dimension')
        return oim[order].transpose((1, 0, 2)).reshape([oim.shape[1], -1])
    else:
        return oim

def reshape(oim, nrois, nlines, order):
    """
    Split and merge a two dimensional plane or three dimensional volume.
    """
    if oim.ndim == 3:
        return array(map(lambda x : merge(split(x, nrois, nlines), order), oim))
    else:
        return merge(split(oim, nrois, nlines), order)

def normalize(oim):
    """
    Normalize a three dimensional volume.
    """
    if not oim.ndim == 3:
        raise Exception('image must be three dimensional')
    means = oim.mean(axis=(1, 2), dtype='float32')
    maximum = means.max()
    return array([oim[i]*maximum/means[i] for i in range(oim.shape[0])])


from numpy import array, multiply, argsort
from glob import glob
import json
from os.path import join

def load(path):
    """
    Load metadata.
    """
    if not path.endswith('json'):
        path = join(path, '*.json')
    files = glob(path)
    if len(files) < 1:
        raise Exception('cannot find metadata file at %s' % path)
    with open(files[0]) as fid:
        header = json.load(fid)
    meta = parse(header)
    return meta

def merge(meta):
    """
    Merge metadata.
    """
    meta = meta.copy()
    center = array([x['center'] for x in meta['rois']]).mean(axis=0)
    size = array([x['size'] for x in meta['rois']]).mean(axis=0)
    size[0] = size[0]*meta['nrois']
    npixels = array([x['npixels'] for x in meta['rois']]).mean(axis=0)
    npixels[0] = npixels[0]*meta['nrois']
    depths = array([x['depths'] for x in meta['rois']]).mean(axis=0)
    meta['rois'] = [dict([('center', list(center)),
        ('size', list(size)),
        ('npixels',list(npixels)),
        ('depths', depths)])]        
    meta['nrois'] = 1
    
    del meta['merge']
    del meta['order']
    del meta['nlines']
    
    return meta

def parse(header):
    """
    Parse metadata from header dictionary.
    """
    meta = {}
    
    if not header['hRoiManager']['mroiEnable']:
        raise Exception('missing mroiEnable field')

    objective_resolution = header['objectiveResolution']/1000

    if header['hFastZ']['enable']:
        depths = header['hFastZ']['userZs']
    else:
        depths = header['hStackManager']['zs']
    
    if type(depths) == int:
        meta['nplanes'] = 1
    else:
        meta['nplanes'] = len(depths)

    meta['depths'] = depths
    
    meta['averaging'] = header['hStackManager']['framesPerSlice']
    meta['power'] = header['hBeams']['powers']

    meta['volumeRate'] = header['hRoiManager']['scanFrameRate']

    meta['rois'] = [dict([('center', rescale(roi['scanfields']['centerXY'], objective_resolution)),
            ('size', rescale(roi['scanfields']['sizeXY'], objective_resolution)),
            ('npixels',roi['scanfields']['pixelResolutionXY']),
            ('depths', roi['zs'])]) 
             for roi in header['imagingRoiGroup']['rois']]
    meta['nrois'] = len(meta['rois'])
    meta['nlines'] = meta['rois'][0]['npixels'][1]

    if type(header['hChannels']['channelSave']) == int:
        meta['nchannels'] = 1
    else:
        meta['nchannels'] = len(header['hChannels']['channelSave'])
    
    meta['order'] = argsort([roi['center'][0] for roi in meta['rois']])
    meta['merge'] = check_order(meta['rois'], meta['order'])
    
    return meta

def check_order(rois, order):
    """
    Check order.
    """
    diff = array([abs((rois[order[i]]['center'][0] + rois[order[i]]['size'][0]/2) 
    - (rois[order[i+1]]['center'][0] - rois[order[i+1]]['size'][0]/2)) for i in range(len(order)-1)])
    
    return diff.mean() < .02

def rescale(x, fovmm):
    """
    Rescale volume.
    """
    return multiply(x, fovmm).tolist()



import os
import json
import click
from os import mkdir
from numpy import inf
from glob import glob
from shutil import rmtree
from os.path import join, isdir
from .. import load, convert

@click.option('--overwrite', is_flag=True, help='Overwrite if directory already exists')
@click.argument('output', nargs=1, metavar='<output directory>', required=False, default=None)
@click.argument('input', nargs=1, metavar='<input directory>', required=True)
@click.command('convert', short_help='process raw data by converting into images', options_metavar='<options>')
def convert_command(input, output, overwrite):
    output = input + '_converted' if output is None else output
    status('reading data from %s' % input)
    status('writing data to %s' % output)
    if isdir(output) and not overwrite:
        error('directory already exists and overwrite is false')
        return
    elif isdir(output) and overwrite:
        rmtree(output)
        mkdir(output)
    if len(glob(join(input, '*.json'))) == 0:
        error('no json metadata found in %s' % input)
        return
    if len(glob(join(input, '*.tif'))) == 0:
        error('no tif or tiff files found in %s' % input)
        return
    data, meta = load(input)
    newdata, newmeta = convert(data, meta)
    newdata.clip(0, inf).astype('uint16').totif(output, overwrite=overwrite)
    with open(join(output, 'metadata.json'), 'w') as f:
      f.write(json.dumps(newmeta))
    success('data written')

def success(msg):
    click.echo('[' + click.style('success', fg='green') + '] ' + msg)

def status(msg):
    click.echo('[' + click.style('convert', fg='blue') + '] ' + msg)

def error(msg):
    click.echo('[' + click.style('error', fg='red') + '] ' + msg)



import json
from numpy import ndarray, array, concatenate, multiply
from thunder.images import fromtif, fromarray
from os import mkdir, makedirs
from os.path import join, exists
from glob import glob
from .metadata import load as load_metadata
from .data import load as load_data
from .data import reshape
from .metadata import merge

def load(path, engine=None):
    """
    Load raw mesoscope data and metadata.

    Parameters
    ----------
    path : str 
        Location with both metadata (.json) and image files (.tif).

    engine : multiple
        A computational backend for parallelization can be None (for local compute) 
        or a SparkContext (for Spark).
    """
    metadata = load_metadata(path)
    data = load_data(path, nplanes=metadata['nplanes'], engine=engine)
    metadata['shape'] = data.shape
    return data, metadata

def convert(data, metadata):
    """
    Convert raw mesoscope data by reshaping and merging rois.

    Parameters
    ----------
    data : numpy array or images
        Raw image data as a numpy array or thunder images object.

    metadata : dict
        Dictionary with metadata.
    """
    if isinstance(data, ndarray):
        data = fromarray(data)
    nchannels = metadata['nchannels']
    nrois = metadata['nrois']
    nlines = metadata['nlines']
    order = metadata['order']
    newdata = data.map(lambda x : reshape(x, nrois, nlines, order))
    newmetadata = merge(metadata)
    newmetadata['shape'] = newdata.shape
    return newdata, newmetadata


from numpy import allclose, inf
from mesoscope import load, convert
from thunder.images import fromtif

def test_load_meta():
    data, meta = load('test/resources/input')
    assert meta['nrois'] == 9
    assert len(meta['rois']) == 9
    assert meta['nplanes'] == 1
    assert meta['nchannels'] == 1
    assert meta['shape'] == (23, 5152, 64)

def test_load_data():
    data, meta = load('test/resources/input')
    assert data.shape == (23, 5152, 64)

def test_convert_meta():
    data, meta = load('test/resources/input')
    newdata, newmeta = convert(data, meta)
    assert newmeta['nrois'] == 1
    assert len(newmeta['rois']) == 1
    assert newmeta['nplanes'] == 1
    assert newmeta['nchannels'] == 1
    assert newmeta['shape'] == (23, 464, 576)

def test_convert_data():
    data, meta = load('test/resources/input')
    newdata, newmeta = convert(data, meta)
    assert newdata.shape == (23, 464, 576)
    assert newdata.dtype == 'int16'

def test_convert_data_array():
    data, meta = load('test/resources/input')
    newdata, newmeta = convert(data.toarray(), meta)
    assert newdata.shape == (23, 464, 576)

def test_convert_ground_truth():
    data, meta = load('test/resources/input')
    newdata, newmeta = convert(data, meta)
    truth = fromtif('test/resources/output')
    assert newdata.shape == truth.shape
    assert allclose(newdata.clip(0, inf), truth)



import click

settings = dict(help_option_names=['-h', '--help'])
from .commands.convert import convert_command

@click.group(options_metavar='', subcommand_metavar='<command>', context_settings=settings)
def cli():
    """
    This is a tool for working with data from the two-photon random access mesoscope. 
    Check out the list of commands to see what you can do.
    """

cli.add_command(convert_command)


#!/usr/bin/env python

from setuptools import setup, find_packages

version = '1.0.1'

required = ['click>=6.6', 'thunder-python>=1.4.1']

setup(
    name='mesoscope',
    version=version,
    description='process data from the two-photon random access mesoscope',
    author='sofroniewn',
    author_email='the.freeman.lab@gmail.com',
    url='https://github.com/sofroniewn/mesoscope',
    packages=find_packages(),
    install_requires=required,
    entry_points = {"console_scripts": ['mesoscope = mesoscope.cli:cli']},
    long_description='See ' + 'https://github.com/sofroniewn/mesoscope',
    license='MIT'
)
